/**
 * This code accompanies the paper:
 *
 * "Dense non-rigid shape correspondence using random forests"
 * E.Rodola, S.Rota Bulo, T.Windheuser, M.Vestner, D.Cremers.
 * Proc. CVPR 2014
 *
 * Please cite the paper above if you use the code.
 */
 
#ifndef RANDOMFORESTCORE_H
#define RANDOMFORESTCORE_H

#include <vector>
#include <cassert>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <stdint.h>
#include <boost/cstdint.hpp>

using namespace std;

namespace vision {

// =====================================================================================
//        Class:  TNode
//  Description:
// =====================================================================================
template<class SplitData, class Prediction> class TNode {
public:
	// ====================  LIFECYCLE     =======================================
	TNode(int start, int end) :
		left(NULL), right(NULL), start(start), end(end), depth(0), id(0) {

	}

	~TNode() {
		if (left!=NULL) {
			delete left;
			left=NULL;
		}
		if (right!=NULL) {
			delete right;
			right=NULL;
		}
	}

	// ====================  ACCESSORS     =======================================
	bool isLeaf() const {
		return left == NULL;
	}
	int getStart() const {
		return start;
	}
	int getEnd() const {
		return end;
	}

	int getNSamples() const {
		return end - start;
	}

	int getDepth() const {
		return depth;
	}

	const SplitData &getSplitData() const {
		return splitData;
	}
	const Prediction &getPrediction() const {
		return prediction;
	}

	const TNode<SplitData, Prediction>* getLeft() const {
		return left;
	}

	const TNode<SplitData, Prediction>* getRight() const {
		return right;
	}

	TNode<SplitData, Prediction>* getLeft() {
		return left;
	}

	TNode<SplitData, Prediction>* getRight() {
		return right;
	}

	uint32_t getID() {
		return id;
	}

	// ====================  MUTATORS      =======================================

	void setSplitData(SplitData splitData) {
		this->splitData = splitData;
	}
	void setPrediction(Prediction prediction) {
		this->prediction = prediction;
	}

	void setDepth(uint16_t depth) {
		this->depth = depth;
	}

	void setEnd(uint32_t end) {
		this->end = end;
	}

	void setStart(uint32_t start) {
		this->start = start;
	}

	void setID(uint32_t id) {
		this->id = id;
	}

	void split(uint32_t middle) {
		assert(middle >= start && middle <= end);
		if (left == NULL) {
			left = new TNode<SplitData, Prediction>(start, middle);
			right = new TNode<SplitData, Prediction>(middle, end);
			left->setDepth(depth + 1);
			right->setDepth(depth + 1);
		} else {
			left->setEnd(middle);
			right->setStart(middle);
		}
	}

	// ====================  OPERATORS     =======================================

protected:
	// ====================  METHODS       =======================================

	// ====================  DATA MEMBERS  =======================================

private:
	// ====================  METHODS       =======================================

	// ====================  DATA MEMBERS  =======================================

	TNode<SplitData, Prediction> *left, *right;
	SplitData splitData;
	Prediction prediction;
	uint32_t start, end;
	uint16_t depth;

	uint32_t id;

};
// -----  end of class TNode  -----


template<class Sample, class Label> struct LabelledSample {
	Sample sample;
	Label label;
};

enum SplitResult
{
  SR_LEFT = 0, SR_RIGHT = 1
};

typedef vector<bool> SplitResultsVector;

// =====================================================================================
//        Class:  RandomTree
//  Description:
// =====================================================================================
template<class SplitData, class Sample, class Label, class Prediction> class RandomTree {
public:
	typedef LabelledSample<Sample, Label> LSample;
	typedef vector<LSample> LSamplesVector;
	typedef TNode<SplitData,Prediction> RandomTreeNode;
	// ====================  LIFECYCLE     =======================================
	RandomTree() :
		root(NULL){
	}

	virtual ~RandomTree() {
		if(root !=NULL)
			delete root;
		root = NULL;
	}

	// ====================  ACCESSORS     =======================================

   void save(string filename, bool includeSamples = false, bool includeNodePredictions = false) const {
		ofstream out(filename.c_str());
		writeHeader(out);
		out << endl;
		write(root, out, includeNodePredictions);
		out << includeSamples << " ";
		if (includeSamples)
			write(this->samples, out);
	}

   void save_binary(string filename, bool includeSamples = false, bool includeNodePredictions = false) const {
      ofstream out(filename.c_str(),std::ios::binary);
      writeHeader_binary(out);
      write_binary(root, out, includeNodePredictions);
      boost::int32_t i = static_cast<boost::int32_t>(includeSamples);
      out.write((const char*)&i, 4);
      if (includeSamples)
         write_binary(this->samples, out);
   }

	Prediction predict(const Sample &sample) const
  {
		assert(root != NULL);
		const RandomTreeNode *curNode = root;
		while (!curNode->isLeaf())
			if (split(curNode->getSplitData(), sample))
				curNode = curNode->getRight();
			else
				curNode = curNode->getLeft();
		return curNode->getPrediction();

	}

	// ====================  MUTATORS      =======================================

	/* BREADTH-FIRST */

	void train(const LSamplesVector trainingSamples) {
		uint32_t nodeId = 0;

		samples.clear();
		samples.resize(trainingSamples.size());
		splitResults.resize(trainingSamples.size());
		copy(trainingSamples.begin(), trainingSamples.end(), samples.begin());
		root = new TNode<SplitData, Prediction>(0, (int) samples.size());
		Prediction rootPrediction;
		initialize(root, rootPrediction);
		root->setPrediction(rootPrediction);
		root->setID(nodeId++);

		vector<TNode<SplitData, Prediction> *> nodeList[2], *curNodeList;
		nodeList[0].reserve(samples.size());
		nodeList[1].reserve(samples.size());

		int nodeListPtr = 0;
		curNodeList = &nodeList[nodeListPtr];

		curNodeList->push_back(root);

		Prediction leftPred, rightPred;
		SplitData splitData;

		while (curNodeList->size() > 0) 
    {
			for (size_t i = 0; i < curNodeList->size(); ++i) 
      {
				RandomTreeNode *node = (*curNodeList)[i];
				if (split(node, splitData, leftPred, rightPred)) 
        {
					int middle;

					doSplit(node, middle);
					node->setSplitData(splitData);
					node->split(middle);
					node->getLeft()->setPrediction(leftPred);
					node->getLeft()->setID(nodeId++);
					node->getRight()->setPrediction(rightPred);
					node->getRight()->setID(nodeId++);
				}
			}

			int nextList = (++nodeListPtr) % 2;
			nodeList[nextList].clear();
			for (size_t i = 0; i < curNodeList->size(); ++i) 
      {
				RandomTreeNode *node = (*curNodeList)[i];
				if (!node->isLeaf()) {
					nodeList[nextList].push_back(node->getLeft());
					nodeList[nextList].push_back(node->getRight());
				}
			}
      
      if(nodeList[nextList].size())
      {
        //cout << nodeList[nextList].size() << " nodes in level " << nodeList[nextList][0]->getDepth() << endl;
      }

			nodeListPtr = nextList;
			curNodeList = &nodeList[nextList];
		}
	}

	void load(string filename) {
		ifstream in(filename.c_str());
		readHeader(in);
		root = new TNode<SplitData, Prediction>(0, 0);
		read(root, in);
		bool includeSamples;
		in >> includeSamples;
		if (includeSamples)
			read(this->samples, in);
	}

   void load_binary(string filename) {
      ifstream in(filename.c_str(), std::ios::binary);
      readHeader_binary(in);
      root = new TNode<SplitData, Prediction>(0, 0);
      read_binary(root, in);
      boost::int32_t includeSamples;
      in.read((char*)&includeSamples, 4);
      if (includeSamples)
         read_binary(this->samples, in);
   }

	// ====================  OPERATORS     =======================================

protected:
	// ====================  METHODS       =======================================

	virtual bool split(const SplitData &splitData, const Sample &sample) const = 0;

	virtual bool split(const RandomTreeNode *node, SplitData &splitData,
			Prediction &leftPrediction, Prediction &rightPrediction) = 0;

	virtual void
			initialize(const RandomTreeNode *node, Prediction &prediction) const = 0;

	const LSamplesVector &getLSamples() const {
		return samples;
	}

	SplitResultsVector &getSplitResults() {
		return splitResults;
	}

	const SplitResultsVector &getSplitResults() const {
		return splitResults;
	}

	virtual void writeHeader(ostream &out) const=0;
	virtual void readHeader(istream &in) =0;

	virtual void write(const Sample &sample, ostream &out) const =0;
	virtual void read(Sample &sample, istream &in) const =0;

	virtual void write(const Prediction &prediction, ostream &out) const=0;
	virtual void read(Prediction &prediction, istream &in) const=0;

	virtual void write(const Label &label, ostream &out) const=0;
	virtual void read(Label &label, istream &in) const=0;

	virtual void write(const SplitData &splitData, ostream &out) const=0;
	virtual void read(SplitData &splitData, istream &in) const=0;

   virtual void writeHeader_binary(ostream &out) const=0;
   virtual void readHeader_binary(istream &in) =0;

   virtual void write_binary(const Sample &sample, ostream &out) const =0;
   virtual void read_binary(Sample &sample, istream &in) const =0;

   virtual void write_binary(const Prediction &prediction, ostream &out) const=0;
   virtual void read_binary(Prediction &prediction, istream &in) const=0;

   virtual void write_binary(const Label &label, ostream &out) const=0;
   virtual void read_binary(Label &label, istream &in) const=0;

   virtual void write_binary(const SplitData &splitData, ostream &out) const=0;
   virtual void read_binary(SplitData &splitData, istream &in) const=0;

	// ====================  DATA MEMBERS  =======================================

private:
	// ====================  METHODS       =======================================

	void doSplit(TNode<SplitData, Prediction> *node, int &pLeft) {
		int pRight = node->getEnd() - 1;
		pLeft = node->getStart();
		LSample s;
		while (pLeft <= pRight) {
			if (splitResults[pLeft]) {
				s = samples[pRight];
				samples[pRight] = samples[pLeft];
				samples[pLeft] = s;
				splitResults[pLeft] = splitResults[pRight];
				splitResults[pRight] = true; //not necessary
				--pRight;
			} else
				++pLeft;
		}
	}

   void write(const TNode<SplitData, Prediction> *node, ostream &out, bool includeNodePrediction = false) const {
		out << (node->isLeaf() ? "L " : "N ");
		out << node->getStart() << " " << node->getEnd() << " ";
		if(!node->isLeaf())
    		{
      			write(node->getSplitData(), out);
		  	out << " ";
    		}
		if(includeNodePrediction || node->isLeaf()){		
			write(node->getPrediction(), out);
		}else
		        write(Prediction(), out);
		out << endl;
		if (!node->isLeaf()) {
			write(node->getLeft(), out);
			write(node->getRight(), out);
		}
	}

   void write_binary(const TNode<SplitData, Prediction> *node, ostream &out, bool includeNodePrediction = false) const {
      boost::int32_t isLeaf = node->isLeaf();
      out.write((const char*)&isLeaf,4);
      boost::int32_t getStart = node->getStart();
      boost::int32_t getEnd = node->getEnd();
      out.write((const char*)&getStart,4);
      out.write((const char*)&getEnd,4);
      if(!node->isLeaf())
         write_binary(node->getSplitData(), out);
      if(includeNodePrediction || node->isLeaf()){
         write_binary(node->getPrediction(), out);
      }else
         write_binary(Prediction(), out);
      if (!node->isLeaf()) {
         write_binary(node->getLeft(), out);
         write_binary(node->getRight(), out);
      }
   }

	void read(TNode<SplitData, Prediction> *node, istream &in) const {
		char type;
		in >> type;
		bool isLeaf = type == 'L';

		int start, end;
		in >> start;
		in >> end;
		
		node->setStart(start);
		node->setEnd(end);

    		if(!isLeaf)
    		{
		  SplitData splitData;
		  read(splitData, in);
		  node->setSplitData(splitData);
    		}		
		Prediction prediction;
		read(prediction, in);
		node->setPrediction(prediction);

		if (!isLeaf) {
			node->split(node->getStart());
			read(node->getLeft(), in);
			read(node->getRight(), in);
		}
	}

   void read_binary(TNode<SplitData, Prediction> *node, istream &in) const {
      boost::int32_t isLeaf;
      in.read((char*)&isLeaf,4);

      boost::int32_t start, end;
      in.read((char*)&start,4);
      in.read((char*)&end,4);

      node->setStart(start);
      node->setEnd(end);

      if(!isLeaf)
      {
         SplitData splitData;
         read_binary(splitData, in);
         node->setSplitData(splitData);
      }
      Prediction prediction;
      read_binary(prediction, in);
      node->setPrediction(prediction);

      if (!isLeaf) {
         node->split(node->getStart());
         read_binary(node->getLeft(), in);
         read_binary(node->getRight(), in);
      }
   }

	void write(const LSamplesVector &lSamples, ostream &out) const {
		out << lSamples.size() << " ";
		for (int i = 0; i < lSamples.size(); ++i) {
			write(lSamples[i].sample, out);
			out << " ";
			write(lSamples[i].label, out);
			out << " ";
		}
	}

   void write_binary(const LSamplesVector &lSamples, ostream &out) const {
      boost::uint32_t sz = lSamples.size();
      out.write((const char*)&sz,4);
      for (int i = 0; i < lSamples.size(); ++i) {
         write_binary(lSamples[i].sample, out);
         write_binary(lSamples[i].label, out);
      }
   }

	void read(LSamplesVector &lSamples, istream &in) const {
		int nSamples;
		in >> nSamples;
		lSamples.resize(nSamples);
		for (int i = 0; i < nSamples; ++i) {
			read(lSamples[i].sample, in);
			read(lSamples[i].label, in);
		}
	}

   void read_binary(LSamplesVector &lSamples, istream &in) const {
      boost::uint32_t nSamples;
      in.read((char*)&nSamples,4);
      lSamples.resize(nSamples);
      for (int i = 0; i < nSamples; ++i) {
         read_binary(lSamples[i].sample, in);
         read_binary(lSamples[i].label, in);
      }
   }

	// ====================  DATA MEMBERS  =======================================

	TNode<SplitData, Prediction> *root;
	LSamplesVector samples;
	SplitResultsVector splitResults;

};
// -----  end of class RandomTree  -----

}
#endif
